package dk.apaq.framework.repository.vaadin;

import com.vaadin.data.Buffered;
import com.vaadin.data.Buffered.SourceException;
import com.vaadin.data.Container;
import com.vaadin.data.Item.PropertySetChangeNotifier;
import com.vaadin.data.Item;
import com.vaadin.data.Property;
import com.vaadin.data.Property.ValueChangeEvent;
import com.vaadin.data.Property.ValueChangeListener;
import com.vaadin.data.Validator.InvalidValueException;
import com.vaadin.data.util.BeanItem;
import dk.apaq.framework.repository.Repository;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

public class RepositoryItem<BT, IDT> implements HasBean<BT>, Item, Buffered, PropertySetChangeNotifier {

    private IDT id;
    private BeanItem<BT> innerItem;
    private final Repository<BT, IDT> repository;
    private final RepositoryContainer<BT, IDT> container;
    private final List<Item.PropertySetChangeListener> listeners = new ArrayList<Item.PropertySetChangeListener>();
    private boolean writeThrough = false;
    private boolean readThrough = false;
    private boolean modified = false;
    private ChangeHandler changeHandler = new ChangeHandler();
    

    private class ChangeHandler implements ValueChangeListener {

        @Override
        public void valueChange(ValueChangeEvent event) {
            modified = true;
            if(writeThrough) {
                commit();
            }
        }
    }

    public RepositoryItem(RepositoryContainer<BT, IDT> container, Repository<BT, IDT> crud, IDT id, BT bean) {
        if(container == null) {
            throw new NullPointerException("Container is null.");
        }
        if(crud == null) {
            throw new NullPointerException("Crud is null.");
        }
        if(bean == null) {
            throw new NullPointerException("Bean is null.");
        }
        this.id = id;
        this.setInnerBean(bean);
        this.repository = crud;
        this.container = container;
    }

    private void reload() {
        if(this.innerItem == null) {
            return;
        }
        BT bean = repository.findOne(id);
        setInnerBean(bean);
    }

    protected BT getInnerBean() {
        return this.innerItem.getBean();
    }

    protected void setInnerBean(BT bean) {
        if (this.innerItem != null) {
            for(Item.PropertySetChangeListener l : listeners) {
                this.innerItem.removeListener(l);
            }

            for(Object propId : this.getItemPropertyIds()) {
                Property prop = this.innerItem.getItemProperty(propId);
                if(prop instanceof Property.ValueChangeNotifier) {
                    ((Property.ValueChangeNotifier)prop).removeListener(changeHandler);
                }
            }
        }

        this.innerItem = new BeanItem<BT>(bean);
        modified = false;

        PropertySetChangeEvent e = new PropertySetChangeEvent() {
            @Override
            public Item getItem() { return RepositoryItem.this; }
        };

        if (this.innerItem != null) {

            //Eventlisteners may add og remove themselves from the list. A copy of the list
            //will let us handle that kind of concurrency.
            List<PropertySetChangeListener> copiedListeners = new ArrayList(listeners);
            for(PropertySetChangeListener l : copiedListeners) {
                l.itemPropertySetChange(e);
            }

            for(PropertySetChangeListener l : listeners) {
                this.innerItem.addListener(l);
            }

            for(Object propId : this.getItemPropertyIds()) {
                Property prop = this.innerItem.getItemProperty(propId);
                if(prop instanceof Property.ValueChangeNotifier) {
                    ((Property.ValueChangeNotifier)prop).addListener(changeHandler);
                }
            }
        }



    }

    @Override
    public BT getBean() {
        return innerItem.getBean();
    }

    @Override
    public Property getItemProperty(Object id) {
        return this.innerItem.getItemProperty(id);
    }

    @Override
    public Collection<?> getItemPropertyIds() {
        return this.innerItem.getItemPropertyIds();
    }

    @Override
    public boolean addItemProperty(Object id, Property property) throws UnsupportedOperationException {
        return this.innerItem.addItemProperty(id, property);
    }

    @Override
    public boolean removeItemProperty(Object id) throws UnsupportedOperationException {
        return this.innerItem.removeItemProperty(id);
    }

    @Override
    public void commit() throws SourceException, InvalidValueException {
        setInnerBean(repository.save(innerItem.getBean()));
        modified = false;
    }

    @Override
    public void discard() throws SourceException {
        if (!modified) {
            return;
        }

        reload();
    }

    @Override
    public boolean isWriteThrough() {
        return writeThrough;
    }

    @Override
    public void setWriteThrough(boolean writeThrough) throws SourceException, InvalidValueException {
        this.writeThrough = writeThrough;
    }

    @Override
    public boolean isReadThrough() {
        return readThrough;
    }

    @Override
    public void setReadThrough(boolean readThrough) throws SourceException {
        this.readThrough = readThrough;
    }

    @Override
    public boolean isModified() {
        return modified;
    }

    @Override
    public void addListener(Item.PropertySetChangeListener listener) {
        innerItem.addListener(listener);
        this.listeners.add(listener);
    }

    
    @Override
    public void removeListener(Item.PropertySetChangeListener listener) {
        innerItem.removeListener(listener);
        this.listeners.add(listener);
    }

}

